#!/usr/bin/env python

# Copyright 2012 Craig Campbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import operator, os

class Util(object):

    @staticmethod
    def arrayDiff(list1, list2):
        if type(list1) is dict:
            list1 = list1.values()
        
        if type(list2) is dict:
            list2 = list2.values()
        
        diff = set(list1) - set(list2)
        
        return list(diff)

    @staticmethod
    def openFile(path):
        file = open(path, "r")
        content = file.read()
        file.close()
        return content

    @staticmethod
    def writeFile(path, content):
        file = open(path, "w")
        file.write(content)
        file.close()

    @staticmethod
    def removeFile(path):
        os.unlink(path)

    @staticmethod
    def sortDictionary(dictionary):
        return sorted(dictionary.iteritems(), key=operator.itemgetter(1))

    @staticmethod
    def isFileOrDirectory(path):
        return os.path.isfile(path) or os.path.isdir(path)



#!/usr/bin/env python

# Copyright 2012 Craig Campbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
class Logger(object):
    BLUE = 'blue'
    LIGHT_BLUE = 'light_blue'
    PINK = 'pink'
    YELLOW = 'yellow'
    WHITE = 'white'
    GREEN = 'green'
    RED = 'red'
    GRAY = 'gray'
    UNDERLINE = 'underline'
    ITALIC = 'italic'

    def __init__(self, options):
        self.verbose = options["verbose"]

    def color(self, message, color, bold = False):
        colors = {
            'italic': '\033[3m',
            'underline': '\033[4m',
            'light_blue': '\033[96m',
            'pink': '\033[95m',
            'blue': '\033[94m',
            'yellow': '\033[93m',
            'green': '\033[92m',
            'red': '\033[91m',
            'gray': '\033[90m',
            'white': '\033[0m'
        }

        end = '\033[0m'

        if not color in colors:
            return message

        bold_start = '\033[1m' if bold else ''

        return colors[color] + bold_start + message + end

    def log(self, message, verbose_only=False):
        if verbose_only and not self.verbose:
            return

        print message



#!/usr/bin/env python

# Copyright 2012 Craig Campbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re, os, subprocess, sys, getopt, shutil
from warpwhistle import WarpWhistle
from util import Util
from instrument import Instrument
from listener import Listener
from logger import Logger

class MusicBox(object):
    VERSION = '1.0.2'

    def processArgs(self, args, local):
        options = {
            'verbose': False,
            'open_nsf': False,
            'listen': False,
            'local': local,
            'create_nsf': True,
            'create_mml': False,
            'separate_voices': False,
            'start': None,
            'end': None
        }

        if '--help' in args:
            return self.showUsage()

        try:
            for key,arg in enumerate(args):
                if arg == '--verbose':
                    options['verbose'] = True
                elif arg == '--open-nsf':
                    options['open_nsf'] = True
                elif arg == '--bob-omb':
                    options['separate_voices'] = True
                elif arg == '--create-nsf':
                    value = args[key + 1]
                    del(args[key + 1])
                    options['create_nsf'] = False if value == '0' else True
                elif arg == '--create-mml':
                    value = args[key + 1]
                    del(args[key + 1])
                    options['create_mml'] = True if value == '1' else False
                elif arg == '--watch':
                    options['listen'] = True
                    value = args[key + 1]
                    del(args[key + 1])
                    bits = value.split(':')
                    options['start'] = bits[0]
                    options['end'] = bits[1] if len(bits) > 1 else None
                else:
                    if options['start'] is None:
                        options['start'] = arg
                    elif options['end'] is None:
                        options['end'] = arg
        except:
            self.showUsage()

        if not options['create_nsf'] and not options['create_mml']:
            self.showUsage('You need to create an MML file or an NSF file')

        if options['start'] is None:
            self.showUsage('You haven\'t specified a file or directory to convert')

        if not Util.isFileOrDirectory(options['start']):
            self.showUsage(self.logger.color(options['start'], self.logger.YELLOW) + " is not a file or directory")

        if options['end'] is None:
            options['end'] = options['start'].replace('.mmlx', '.mml')

        options['end'] = options['end'].replace('.nsf', '.mml')

        return options

    def play(self, args, local):
        # create an intial logger so we can log before args are processed
        self.logger = Logger({"verbose": False})
        self.drawLogo()

        options = self.processArgs(args, local)
        self.options = options

        # set up the logger with the correct options
        self.logger = Logger(self.options)

        listener = Listener(self.logger)
        listener.onChange(self.processFile)

        if os.path.isdir(options['start']) and not os.path.isdir(options['end']):
            os.mkdir(options['end'])

        if self.options['listen']:
            listener.watch(options['start'], options['end'])
        else:
            listener.process(options['start'], options['end'])
            self.logger.log(self.logger.color('Done!', self.logger.PINK))
            sys.exit(0)

    def drawLogo(self):
        self.logger.log(self.logger.color('_|      _|  _|      _|  _|        _|      _|  ', self.logger.PINK))
        self.logger.log(self.logger.color('_|_|  _|_|  _|_|  _|_|  _|          _|  _|    ', self.logger.PINK))
        self.logger.log(self.logger.color('_|  _|  _|  _|  _|  _|  _|            _|      ', self.logger.PINK))
        self.logger.log(self.logger.color('_|      _|  _|      _|  _|          _|  _|    ', self.logger.PINK))
        self.logger.log(self.logger.color('_|      _|  _|      _|  _|_|_|_|  _|      _|  ', self.logger.PINK))
        self.logger.log(self.logger.color('                                 version ' + self.logger.color(MusicBox.VERSION, self.logger.PINK, True) + '\n', self.logger.PINK))

    def showUsage(self, message = None):
        logger = self.logger
        if message is not None:
            logger.log(logger.color('ERROR:', logger.RED))
            logger.log(message + "\n")

        logger.log(logger.color('ARGUMENTS:', logger.WHITE, True))
        logger.log(logger.color('--help', logger.WHITE) + '                                shows help dialogue')
        logger.log(logger.color('--verbose', logger.WHITE) + '                             shows verbose output')
        logger.log(logger.color('--open-nsf', logger.WHITE) + '                            opens nsf file on save')
        logger.log(logger.color('--bob-omb', logger.WHITE) + '                             generates a separate NSF file for each voice')
        logger.log(logger.color('--create-mml ' + logger.color('0', logger.YELLOW), logger.WHITE) +'                        creates an MML file on save (defaults to 0)')
        logger.log(logger.color('--create-nsf ' + logger.color('1', logger.YELLOW), logger.WHITE) +'                        creates an NSF file on save (defaults to 1)')
        logger.log(logger.color('--watch', logger.WHITE) + logger.color(' path/to/mmlx', logger.YELLOW) + logger.color(':', logger.GRAY) + logger.color('path/to/mml', logger.YELLOW) + '      watches for changes in first directory and compiles to second')
        logger.log(logger.color('\nEXAMPLES:', logger.WHITE, True))

        logger.log(logger.color('watch directory for changes in .mmlx files:', logger.GRAY))
        logger.log(logger.color('mmlx --watch path/to/mmlx', logger.BLUE, True))

        logger.log(logger.color('\nkeep compiled files in mml directory and open .nsf file on save:', logger.GRAY))
        logger.log(logger.color('mmlx --watch path/to/mmlx:path/to/mml --open-nsf', logger.BLUE, True))

        logger.log(logger.color('\nwatch a single file for changes:', logger.GRAY))
        logger.log(logger.color('mmlx --watch path/to/file.mmlx:path/to/otherfile.mml', logger.BLUE, True))

        logger.log(logger.color('\nrun once for a single file:', logger.GRAY))
        logger.log(logger.color('mmlx path/to/file.mmlx path/to/otherfile.nsf', logger.BLUE, True))
        logger.log(logger.color('mmlx path/to/file.mmlx', logger.BLUE, True))

        logger.log(logger.color('\nrun once for a directory:', logger.GRAY))
        logger.log(logger.color('mmlx path/to/mmlx path/to/nsf', logger.BLUE, True))
        logger.log('')

        sys.exit(1)

    def createNSF(self, path, open_file = False):
        self.logger.log('generating file: ' + self.logger.color(path.replace('.mml', '.nsf'), self.logger.YELLOW))

        nes_include_path = os.path.join(os.path.dirname(__file__), 'nes_include')
        bin_dir = os.path.join(nes_include_path, '../../bin')

        # copy the nes include directory locally so we don't have to deal with permission problems
        if not self.options['local']:
            new_nes_include_path = os.path.join(os.path.dirname(path), 'nes_include')
            shutil.copytree(nes_include_path, new_nes_include_path)
            nes_include_path = new_nes_include_path

        os.environ['NES_INCLUDE'] = nes_include_path
        command = os.path.join(bin_dir, 'ppmckc') if self.options['local'] else 'ppmckc'
        output = subprocess.Popen([command, '-m1', '-i', path], stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        parts = output.communicate()
        stdout = parts[0]
        stderr = parts[1]

        command = os.path.join(bin_dir, 'nesasm') if self.options['local'] else 'nesasm'
        output = subprocess.Popen([command, '-s', '-raw', os.path.join(nes_include_path, 'ppmck.asm')], stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        parts = output.communicate()
        stdout = parts[0]
        stderr = parts[1]

        nsf_path = os.path.join(nes_include_path, 'ppmck.nes')
        if not os.path.isfile(nsf_path):
            os.unlink('define.inc')
            self.logger.log('')
            if not self.options['local']:
                shutil.rmtree(nes_include_path)
            raise Exception('failed to create NSF file! Your MML is probably invalid.')

        os.rename(nsf_path, path.replace('.mml', '.nsf'))
        os.unlink('define.inc')
        os.unlink('effect.h')
        os.unlink(path.replace('.mml', '.h'))
        if not self.options['local']:
            shutil.rmtree(nes_include_path)

        if open_file and self.options['open_nsf']:
            subprocess.call(['open', path.replace('.mml', '.nsf')])

    def processFile(self, input, output, open_file = False):
        Instrument.reset()

        self.logger.log('processing file: ' + self.logger.color(input, self.logger.YELLOW), True)
        original_content = content = Util.openFile(input)

        whistle = WarpWhistle(content, self.logger, self.options)
        whistle.import_directory = os.path.dirname(input)

        while whistle.isPlaying():
            open_file = open_file and whistle.first_run

            song = whistle.play()

            new_output = output
            if song[1] is not None:
                new_output = new_output.replace('.mml', '_' + song[1] + '.mml')

            self.handleProcessedFile(song[0], new_output, open_file)

        if self.options['separate_voices']:
            self.logger.log("")

    def handleProcessedFile(self, content, output, open_file = False):
        if self.options['create_mml']:
            self.logger.log('generating file: ' + self.logger.color(output, self.logger.YELLOW))

        Util.writeFile(output, content)

        if self.options['create_nsf']:
            self.createNSF(output, open_file)

        if not self.options['create_mml']:
            Util.removeFile(output)

        if open_file:
            self.logger.log("")



#!/usr/bin/env python

# Copyright 2012 Craig Campbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os, glob, time, sys

class Listener(object):

    def __init__(self, logger=None):
        self.watching = False
        self.callback = None
        self.first_run = True
        self.file_list = {}
        self.logger = logger

    def getFilesFromDir(self, path, extension = ""):
        path = path + "/*"

        if not extension == "":
            path = path + "." + extension.lstrip(".")

        return glob.glob(path)

    def onChange(self, callback):
        self.callback = callback

    def process(self, start, end, is_dir = None):
        try:
            if is_dir is None:
                is_dir = os.path.isdir(start)

            if is_dir:
                files = self.getFilesFromDir(start, "mmlx")
                output_file = None
            else:
                files = [start]
                output_file = end

            initial_output_file = output_file

            for file in files:
                output_file = initial_output_file

                filename = os.path.basename(file)
                if filename.startswith('_'):
                    continue

                last_changed = os.stat(file).st_mtime
                output_file = output_file if output_file is not None else os.path.join(end, filename.replace(".mmlx", ".mml"))

                if not self.file_list.has_key(file):
                    self.file_list[file] = last_changed
                    self.callback(file, output_file)
                    continue

                if last_changed != self.file_list[file]:
                    self.logger.log(self.logger.color("detected change to: ", self.logger.GRAY) + self.logger.color(file, self.logger.UNDERLINE))
                    self.file_list[file] = last_changed
                    self.callback(file, output_file, True)

            if self.first_run:
                self.logger.log('')
                self.first_run = False

        except Exception:
            self.logger.log(self.logger.color('Sorry, an error occured:\n', self.logger.RED))
            import traceback
            lines = traceback.format_exc().splitlines()
            self.logger.log(self.logger.color(lines.pop(), self.logger.RED) + '\n')
            self.logger.log('\n'.join(lines))
            self.logger.log('')

            # continue watching
            if self.watching:
                self.watch(start, end)
                return

            sys.exit(1)

    def watch(self, start, end):
        try:
            self.watching = True
            is_dir = os.path.isdir(start)
            while 1:
                self.process(start, end, is_dir)
                time.sleep(.5)
        except KeyboardInterrupt:
            phrases = [
                'Sayonara!',
                'Goodbye! Have a nice day!',
                'Come back soon!',
                'Sad to see you leaving already!',
                'Hasta la vista, baby'
            ]
            from random import choice
            self.logger.log(self.logger.color('\n' + choice(phrases), self.logger.PINK))
            sys.exit(0)



#!/usr/bin/env python

# Copyright 2012 Craig Campbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from util import Util
import math

class Instrument(object):

    def __init__(self, data):
        valid_chips = ['N106', 'FDS', 'VRC6', 'VRC7']

        for key in data:
            if key == 'chip' and data[key] not in valid_chips:
                raise Exception('value for chip is not valid: ' + data[key])

            setattr(self, key, data[key])

        if hasattr(self, "adsr"):
            self.volume = self.getVolumeFromADSR(self.adsr)
        
        if self.getChip() == 'N106':
            Instrument.N106_buffers[self.waveform] = int(self.buffer) if hasattr(self, 'buffer') else None

    # attack - time taken for amplitude to rise from 0 to max (15)
    # decay - time taken for amplitude to drop to sustain level
    # sustain - amplitude at which the note is held
    # release - time taken for amplitude to drop from sustain level to 0
    #
    # note:
    # if decay is 0, max amplitude is the sustain value
    #
    def getVolumeFromADSR(self, adsr):
        bits = adsr.split(' ')
        attack = bits[0]
        decay = bits[1]
        sustain = bits[2]
        release = bits[3]

        max_volume = sustain if int(decay) == 0 else 15

        if hasattr(self, "max_volume"):
            max_volume = int(self.max_volume)

        if attack == '0' and decay == '0' and sustain == '0' and release != '0':
            sustain = 15

        volume = ''
        if attack != '0':
            volume += str(max_volume) + ' ' if attack == '0' else ' '.join(self.divideIntoSteps(0, max_volume, attack)) + ' '

        if sustain != '0':
            volume += ' '.join(self.divideIntoSteps(max_volume, sustain, decay)) + ' '

        if release != '0':
            volume += ' '.join(self.divideIntoSteps(sustain, 0, release)) + ' '

        if volume.strip() == '':
            volume = '0'

        return volume

    def divideIntoSteps(self, min, max, steps):
        min = int(min)
        max = int(max)
        steps = int(steps)

        # print 'MIN', min
        # print 'MAX', max
        # print 'STEPS', steps

        if steps == 1:
            return [str(min), str(max)]

        # figure out the equation of a line to match these coordinates
        # coordinates are (0, min) and (steps - 1, max)
        slope = float(max - min) / float(steps - 1)
        equation = lambda x: slope * x + min

        values = []
        for x in range(0, steps):
            value = int(math.ceil(equation(x)))
            values.append(str(value))

        # print values
        return values

    def getChip(self):
        if hasattr(self, 'chip'):
            return self.chip
        
        return None
        
    @staticmethod
    def reset(counter = 0):
        counter = int(counter)

        Instrument.counters = {
            'timbre': counter,
            'volume': counter,
            'pitch': counter,
            'arpeggio': counter,
            'vibrato': counter,
            'N106': counter,
            'FDS': counter
        }

        Instrument.timbres = {}
        Instrument.volumes = {}
        Instrument.pitches = {}
        Instrument.arpeggios = {}
        Instrument.vibratos = {}
        Instrument.N106 = {}
        Instrument.N106_buffers = {}
        Instrument.FDS = {}

    def hasParent(self):
        return hasattr(self, 'extends') and self.extends is not None

    def inherit(self, instrument):
        dictionary = instrument.__dict__
        for key in dictionary:
            if key == "extends" or not hasattr(self, key):
                setattr(self, key, dictionary[key])

        if not "extends" in dictionary:
            delattr(self, "extends")

    def getParent(self):
        return self.extends

    def getCountFor(self, macro):
        i = Instrument.counters[macro]
        Instrument.counters[macro] += 1
        return i

    def getVolumeMacro(self):
        i = Instrument.volumes[self.volume] if self.volume in Instrument.volumes else self.getCountFor('volume')
        Instrument.volumes[self.volume] = i
        return '@v' + str(i)

    def getPitchMacro(self):
        i = Instrument.pitches[self.pitch] if self.pitch in Instrument.pitches else self.getCountFor('pitch')
        Instrument.pitches[self.pitch] = i
        return 'EP' + str(i)

    def getArpeggioMacro(self):
        i = Instrument.arpeggios[self.arpeggio] if self.arpeggio in Instrument.arpeggios else self.getCountFor('arpeggio')
        Instrument.arpeggios[self.arpeggio] = i
        return 'EN' + str(i)

    def getTimbreMacro(self):
        i = Instrument.timbres[self.timbre] if self.timbre in Instrument.timbres else self.getCountFor('timbre')
        Instrument.timbres[self.timbre] = i
        return '@@' + str(i)

    def getVibratoMacro(self):
        i = Instrument.vibratos[self.vibrato] if self.vibrato in Instrument.vibratos else self.getCountFor('vibrato')
        Instrument.vibratos[self.vibrato] = i
        return 'MP' + str(i)
    
    def getN106Macro(self):
        i = Instrument.N106[self.waveform] if self.waveform in Instrument.N106 else self.getCountFor('N106')
        Instrument.N106[self.waveform] = i
        return '@@' + str(i)
    
    def getFDSMacro(self):
        i = Instrument.FDS[self.waveform] if self.waveform in Instrument.FDS else self.getCountFor('FDS')
        Instrument.FDS[self.waveform] = i
        return '@@' + str(i)
        

    @staticmethod
    def hasBeenUsed():
        macros = ["timbres", "volumes", "pitches", "arpeggios", "vibratos", "N106", "FDS"]
        for macro in macros:
            if len(getattr(Instrument, macro)) > 0:
                return True

        return False

    @staticmethod
    def render():
        macros = ''

        # render timbres
        for timbre in Util.sortDictionary(Instrument.timbres):
            macros += '@' + str(timbre[1]) + ' = { ' + timbre[0] + ' }\n'

        # render volumes
        for volume in Util.sortDictionary(Instrument.volumes):
            macros += '@v' + str(volume[1]) + ' = { ' + volume[0] + ' }\n'

        # render pitches
        for pitch in Util.sortDictionary(Instrument.pitches):
            macros += '@EP' + str(pitch[1]) + ' = { ' + pitch[0] + ' }\n'

        # render arpeggios
        for arpeggio in Util.sortDictionary(Instrument.arpeggios):
            macros += '@EN' + str(arpeggio[1]) + ' = { ' + arpeggio[0] + ' }\n'

        # render vibratos
        for vibrato in Util.sortDictionary(Instrument.vibratos):
            macros += '@MP' + str(vibrato[1]) + ' = { ' + vibrato[0] + ' }\n'
        
        # render N106
        for macro in Util.sortDictionary(Instrument.N106):
            waveform = Instrument.validateN106(macro[0])
            macros += '@N' + str(macro[1]) + ' = { ' + Instrument.getN106Buffer(waveform) + ', ' + waveform + ' }\n'

        # render FDS
        for macro in Util.sortDictionary(Instrument.FDS):
            macros += '@FM' + str(macro[1]) + ' = { ' + Instrument.validateFds(macro[0]) + ' }\n'

        return macros
        
    @staticmethod
    def validateN106(macro):
        bits = macro.strip().split(' ')
        if len(bits) % 4 != 0:
            raise Exception('N106 waveform samples have to be a multiple of 4')
        
        for bit in bits:
            bit = int(bit.replace('$', ''), 16) if bit.startswith('$') else int(bit)
            if bit < 0:
                raise Exception('N106 waveform parameter cannot be less than 0')

            if bit > 15:
                raise Exception('N106 waveform parameter cannot be greater than 15')
        
        return macro
        
    @staticmethod
    def validateFds(macro):
        bits = macro.strip().split(' ')
        if len(bits) != 64:
            raise Exception('FDS waveform must have exactly 64 parameters')
        
        for bit in bits:
            bit = int(bit)
            if bit < 0:
                raise Exception('FDS waveform parameter cannot be less than 0')
            
            if bit > 63:
                raise Exception('FDS waveform parameter cannot be greater than 63')
        
        return macro
    
    @staticmethod
    def maxBufferFromSampleLength(sample_length):
        map = {
            32: 3,
            28: 3,
            24: 4,
            20: 5,
            16: 7,
            12: 9,
             8: 13,
             4: 32
        }

        return map[sample_length]

    @staticmethod
    def getBufferForWaveform(waveform):
        return Instrument.N106_buffers[waveform]

    @staticmethod
    def getN106Buffer(waveform):
        waveform = waveform.strip()
        bits = waveform.split(' ')
            
        max_allowed_buffer = Instrument.maxBufferFromSampleLength(len(bits))
        buffer = Instrument.getBufferForWaveform(waveform)
                
        if buffer is None:
            return '00'

        if buffer > max_allowed_buffer:
            raise Exception('buffer value cannot be greater than: ' + str(max_allowed_buffer) + ' for ' + str(len(bits)) + ' samples')
        
        if buffer < 10:
            buffer = '0' + str(buffer)
        
        return str(buffer)

    def start(self, whistle):
        start = ''
        if hasattr(self, 'timbre'):
            last_timbre = whistle.getDataForVoice(whistle.current_voices[0], 'timbre')
            new_timbre = self.getTimbreMacro()

            if new_timbre != last_timbre:
                whistle.setDataForVoices(whistle.current_voices, 'timbre', new_timbre)
                start += new_timbre + ' '

        if hasattr(self, 'volume'):
            last_volume = whistle.getDataForVoice(whistle.current_voices[0], 'volume')
            new_volume = self.getVolumeMacro()

            if new_volume != last_volume:
                whistle.setDataForVoices(whistle.current_voices, 'volume', new_volume)
                start += new_volume + ' '

        if hasattr(self, 'pitch'):
            last_pitch = whistle.getDataForVoice(whistle.current_voices[0], 'pitch')
            new_pitch = self.getPitchMacro()

            if new_pitch != last_pitch:
                whistle.setDataForVoices(whistle.current_voices, 'pitch', new_pitch)
                start += new_pitch + ' '

        if hasattr(self, 'arpeggio'):
            last_arpeggio = whistle.getDataForVoice(whistle.current_voices[0], 'arpeggio')
            new_arpeggio = self.getArpeggioMacro()

            if new_arpeggio != last_arpeggio:
                whistle.setDataForVoices(whistle.current_voices, 'arpeggio', new_arpeggio)
                start += new_arpeggio + ' '

        if hasattr(self, 'vibrato'):
            last_vibrato = whistle.getDataForVoice(whistle.current_voices[0], 'vibrato')
            new_vibrato = self.getVibratoMacro()

            if new_vibrato != last_vibrato:
                whistle.setDataForVoices(whistle.current_voices, 'vibrato', new_vibrato)
                start += new_vibrato + ' '

        if hasattr(self, 'q'):
            last_q = whistle.getDataForVoice(whistle.current_voices[0], 'q')
            new_q = 'q' + self.q

            if new_q != last_q:
                whistle.setDataForVoices(whistle.current_voices, 'q', new_q)
                start += new_q + ' '
        
        if hasattr(self, 'waveform') and self.getChip() == 'N106':
            last_n106 = whistle.getDataForVoice(whistle.current_voices[0], 'timbre')
            new_n106 = self.getN106Macro()

            if new_n106 != last_n106:
                whistle.setDataForVoices(whistle.current_voices, 'timbre', new_n106)
                start += new_n106 + ' '
        
        if hasattr(self, 'waveform') and self.getChip() == 'FDS':
            last_fds = whistle.getDataForVoice(whistle.current_voices[0], 'timbre')
            new_fds = self.getFDSMacro()

            if new_fds != last_fds:
                whistle.setDataForVoices(whistle.current_voices, 'timbre', new_fds)
                start += new_fds + ' '

        return start

    def end(self, whistle):
        end = ''
        if hasattr(self, 'pitch'):
            whistle.setDataForVoices(whistle.current_voices, 'pitch', None)
            end += 'EPOF '

        if hasattr(self, 'arpeggio'):
            whistle.setDataForVoices(whistle.current_voices, 'arpeggio', None)
            end += 'ENOF '

        if hasattr(self, 'vibrato'):
            whistle.setDataForVoices(whistle.current_voices, 'vibrato', None)
            end += 'MPOF '

        return end



#!/usr/bin/env python

# Copyright 2012 Craig Campbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re, os, math
from util import Util
from instrument import Instrument

class WarpWhistle(object):
    TEMPO = 'tempo'
    VOLUME = 'volume'
    TIMBRE = 'timbre'
    ARPEGGIO = 'arpeggio'
    INSTRUMENT = 'instrument'
    PITCH = 'pitch'
    OCTAVE = 'octave'
    SLIDE = 'slide'
    Q = 'q'
    
    ABSOLUTE_NOTES = 'X-ABSOLUTE-NOTES'
    TRANSPOSE = 'X-TRANSPOSE'
    COUNTER = 'X-COUNTER'
    X_TEMPO = 'X-TEMPO'
    SMOOTH = 'X-SMOOTH'
    N106 = 'EX-NAMCO106'
    FDS = 'EX-DISKFM'
    VRC6 = 'EX-VRC6'
    VRC7 = 'EX-VRC7'
    PITCH_CORRECTION = 'PITCH-CORRECTION'
    
    CHIP_N106 = 'N106'
    CHIP_FDS = 'FDS'
    CHIP_VRC6 = 'VRC6'
    CHIP_VRC7 = 'VRC7'

    def __init__(self, content, logger, options):
        self.first_run = True

        # current voice we are processing if we are processing voices separately
        self.process_voice = None

        # list of voices to process
        self.voices_to_process = None

        # list of voices
        self.voices = None

        self.content = content
        self.logger = logger
        self.options = options
        self.reset()

    def reset(self):
        self.current_voices = []
        self.global_vars = {}
        self.vars = {}
        self.instruments = {}
        self.data = {}
        self.global_lines = []

    def getDataForVoice(self, voice, key):
        if not voice in self.data:
            return None

        if not key in self.data[voice]:
            return None

        return self.data[voice][key]

    def setDataForVoice(self, voice, key, value):
        if not voice in self.data:
            self.data[voice] = {}

        self.data[voice][key] = value

    def setDataForVoices(self, voices, key, value):
        for voice in voices:
            self.setDataForVoice(voice, key, value)

    def getVar(self, key):
        if key in self.vars:
            return self.vars[key]

        return None

    def getGlobalVar(self, key):
        if key in self.global_vars:
            return self.global_vars[key]

        if key == WarpWhistle.TRANSPOSE:
            return 0

        return None

    def processImports(self, content):
        matches = re.findall(r'(@import\s{1,}(\'|\")(.*)(\2))$', content, re.MULTILINE)

        for match in matches:
            filename = match[2] if match[2].endswith('.mmlx') else match[2] + '.mmlx'
            disk_path = os.path.join(self.import_directory, filename)
            file_content = Util.openFile(disk_path)
            content = content.replace(match[0], file_content)

        return content

    def stripComments(self, content):
        # replace all /* comments */
        content = re.sub(re.compile(r'(/\*(.*?)\*/)', re.MULTILINE | re.DOTALL), '', content)

        # replace all ; comments
        content = re.sub(re.compile(r' {0,}(;.*)$', re.MULTILINE), '', content)

        # replace all // comments
        content = re.sub(re.compile(r' {0,}(//.*)$', re.MULTILINE), '', content)

        # replace empty lines
        content = re.sub(re.compile(r'\n{2,}', re.MULTILINE), '\n', content)

        return content

    def collapseSpaces(self, content):
        # collapse multiple spaces into a single space
        return re.sub(re.compile(' {2,}', re.MULTILINE), ' ', content)

    def removeBlankLines(self, content):
        return re.sub(r'\n{2,}', '\n', content)

    def processGlobalVariables(self, content):
        matches = re.findall(r'(^#([-A-Z0-9]+)( {1,}(.*))?\n)', content, re.MULTILINE)
        for match in matches:
            if match[1] == WarpWhistle.TRANSPOSE:
                self.global_vars[match[1]] = int(match[3]) if match[3] else 0
            else:
                self.global_vars[match[1]] = match[3] or True

            if match[1].startswith('X-'):
                content = content.replace(match[0], '')
            else:
                self.global_lines.append(match[0])

            if match[1] == WarpWhistle.COUNTER:
                Instrument.reset(match[3])

        return content

    def isReserved(self, var):
        # single letter match
        if re.match(r'^[A-Z]{1,2}$', var):
            return True

        # volume macro
        if re.match(r'^v\d+$', var):
            return True

        # rest
        if re.match(r'^r\d+$', var):
            return True

        # wait
        if re.match(r'^w\d+$', var):
            return True

        # pitch macro
        if re.match(r'^EP\d+$', var):
            return True

        # arpeggio macro
        if re.match(r'^EN\d+$', var):
            return True

        # self delay macro
        if re.match(r'^SD\d+$', var):
            return True

        reserved = ['EPOF', 'ENOF', 'SDOF', 'w', 'r']

        return var in reserved

    def processLocalVariables(self, content):
        matches = re.findall(r'(^([a-zA-Z]{1}([a-zA-Z0-9_]+)?)\s{0,}=\s{0,}(.*)\n)', content, re.MULTILINE)
        for match in matches:

            if self.isReserved(match[1]):
                raise Exception('variable ' + match[1] + ' is reserved')

            self.vars[match[1]] = match[3]

            content = content.replace(match[0], '')

        return content

    def processVariables(self, content):
        content = self.processGlobalVariables(content)
        content = self.processLocalVariables(content)
        return content

    def addInstrument(self, name, content):
        if name == 'end':
            raise Exception('end is a reserved word and connt be used for an instrument')

        lines = content.strip().split('\n')
        data = {}

        for line in lines:
            line = line.strip()
            match = re.match(r'^@extends {1,}(\'|\")(.*)(\1)$', line)
            if match:
                data["extends"] = match.group(2)
                continue

            data[line.split(':')[0].strip()] = line.split(':')[1].strip()

        self.instruments[name] = Instrument(data)

    def updateInstruments(self):
        for name in self.instruments:
            instrument = self.instruments[name]
            while instrument.hasParent():
                instrument.inherit(self.instruments[instrument.getParent()])

    def processInstruments(self, content):
        matches = re.findall(r'(^([a-zA-Z0-9-_]+):( {0,}(\n( {4}|\t)(.*))+)\n)', content, re.MULTILINE)
        for match in matches:
            self.addInstrument(match[1].lower(), match[2])
            content = content.replace(match[0], '')

        self.updateInstruments()
        return content
    
    def getVoicesForChip(self, chip):
        if chip == WarpWhistle.CHIP_N106:
            return {
                'A': 'P',
                'B': 'Q',
                'C': 'R',
                'D': 'S',
                'E': 'T',
                'F': 'U',
                'G': 'V',
                'H': 'W'
            }
        
        if chip == WarpWhistle.CHIP_FDS:
            return {
                'A': 'F'
            }
        
        if chip == WarpWhistle.CHIP_VRC6:
            return {
                'A': 'M',
                'B': 'N',
                'C': 'O'
            }
        
        if chip == WarpWhistle.CHIP_VRC7:
            return {
                'A': 'G',
                'B': 'H',
                'C': 'I',
                'D': 'J',
                'E': 'K',
                'F': 'L'
            }
    
    def getVoiceTranslation(self, chip, voice):
        voices = self.getVoicesForChip(chip)
        
        if not voice in voices:
            return None
                
        return voices[voice]
    
    def getVoiceFor(self, chip, voices):
        final_voices = []
        voices = list(voices)
        for voice in voices:
            new_voice = self.getVoiceTranslation(chip, voice)
            if new_voice is None:
                raise Exception('voice: ' + voice + ' is outside of the voice range for chip: ' + chip)
            
            final_voices.append(new_voice)
        
        return ''.join(final_voices)
    
    def processExpansionVoices(self, content):
        """finds any special voices (such as N106-AB) and converts them to the proper voice names"""
        matches = re.findall(r'((N106|FDS|VRC6|VRC7)-([A-Z]+) )', content)
        for match in matches:
            content = content.replace(match[0], self.getVoiceFor(match[1], match[2]) + ' ')

        return content
    
    def renderTempo(self, content):
        tempo = self.getGlobalVar(WarpWhistle.X_TEMPO)
        if tempo is None:
            return content
        
        return self.addToMml(content, "".join(self.voices) + " t" + str(tempo) + "\n")

    def renderInstruments(self, content):
        if not Instrument.hasBeenUsed():
            return content

        # find the last #BLOCK on the top of the file and render the instruments below it
        return self.addToMml(content, Instrument.render())

    def renderN106(self, content):
        n106_voices = self.getVoicesForChip(WarpWhistle.CHIP_N106).values()
        n106_voices.sort()
        
        n106_count = 0
        for voice in self.voices:
            if voice in n106_voices:
                index = n106_voices.index(voice) + 1
                n106_count = max(n106_count, index)
            
        if n106_count > 0:
            # in order to stay on pitch it has to be 1,2,4 or 8
            if n106_count == 3:
                n106_count = 4
            
            if n106_count in [5, 6, 7]:
                n106_count = 8
            
            if not WarpWhistle.N106 in self.global_vars:
                self.global_vars[WarpWhistle.N106] = str(n106_count)
                content = self.addToMml(content, '#' + WarpWhistle.N106 + ' ' + str(n106_count) + '\n', True)
            
            if not WarpWhistle.PITCH_CORRECTION in self.global_vars:
                self.global_vars[WarpWhistle.PITCH_CORRECTION] = True
                content = self.addToMml(content, '#' + WarpWhistle.PITCH_CORRECTION + '\n', True)
        
        return content

    def getExpForChip(self, chip):
        if chip == WarpWhistle.CHIP_N106:
            return WarpWhistle.N106
        elif chip == WarpWhistle.CHIP_FDS:
            return WarpWhistle.FDS
        elif chip == WarpWhistle.CHIP_VRC6:
            return WarpWhistle.VRC6
        elif chip == WarpWhistle.CHIP_VRC7:
            return WarpWhistle.VRC7

    def renderForChip(self, chip, content):
        chip_voices = self.getVoicesForChip(chip).values()
        chip_voices.sort()
        
        used = False
        for voice in self.voices:
            if voice in chip_voices:
                used = True
                break
        
        exp = self.getExpForChip(chip)
        if used and not exp in self.global_vars:
            self.global_vars[exp] = True
            content = self.addToMml(content, '#' + exp + '\n', True)
            
        return content
        
    def renderExpansionChips(self, content):
        content = self.renderN106(content)
        content = self.renderForChip(WarpWhistle.CHIP_FDS, content)
        content = self.renderForChip(WarpWhistle.CHIP_VRC6, content)
        content = self.renderForChip(WarpWhistle.CHIP_VRC7, content)
        return content

    def addToMml(self, content, string, is_global_line = False):
        last_global_declaration = self.global_lines[-1]
        
        if is_global_line:
            self.global_lines.append(string)
            
        return content.replace(last_global_declaration, last_global_declaration + string)

    def replaceVariables(self, content):
        for key in self.vars:
            pattern = '((?<=\s)|(?<=\[))' + key + '(?=\s|\Z|\])'
            content = re.sub(re.compile(pattern, re.MULTILINE), self.vars[key], content)

        return content

    def isUndefinedVariable(self, var):
        return False

    # def processVoice(self, voice, content):
    #     regex = '(' + voice + '\s{0,}(.*?)(?=(\n[A-Z^' + voice + ']{1,2}\s|\n?\Z)))'
    #     matches = re.findall(regex, content, re.MULTILINE | re.DOTALL)
    #     print matches
    #     return content

    def moveToOctave(self, new_octave, last_octave):
        diff = new_octave - last_octave
        ticks = abs(diff)
        symbol = '>' if diff > 0 else '<'
        return symbol * ticks

    def getNumberForNote(self):
        return {
            'c': 0,
            'c+': 1,
            'd-': 1,
            'd': 2,
            'd+': 3,
            'e-': 3,
            'e': 4,
            'f-': 4,
            'e+': 5,
            'f': 5,
            'f+': 6,
            'g-': 6,
            'g': 7,
            'g+': 8,
            'a-': 8,
            'a': 9,
            'a+': 10,
            'b-': 10,
            'b': 11
        }

    def getNoteForNumber(self):
        return {
            0: 'c',
            1: 'c+',
            2: 'd',
            3: 'd+',
            4: 'e',
            5: 'f',
            6: 'f+',
            7: 'g',
            8: 'g+',
            9: 'a',
            10: 'a+',
            11: 'b'
        }

    def isNoiseChannel(self):
        return self.current_voices[0] == 'D'

    def getFrequency(self, note, octave):
        frequencies = [
            0x06AE,
            0x064E,
            0x05F4,
            0x059E,
            0x054E,
            0x0501,
            0x04B9,
            0x0476,
            0x0436,
            0x03F9,
            0x03C0,
            0x038A
        ]

        index = self.getNumberForNote()[note]
        frequency = frequencies[index] >> (octave - 2)
        return frequency

    def calculateN106OctaveShift(self, channel_count, waveform):
        if waveform is None:
            return 0

        # my math skills are so lacking these days this is the number of octaves to shift
        # based on the waveform sample count and the channel count
        shift = {
            4: 4,
            8: 3,
            16: 2,
            32: 1,
            64: 0,
            128: -1,
            256: -2
        }

        number = channel_count * len(waveform.strip().split(' '))
        return shift[number]

    def slide(self, start_data, end_data):
        N106_channels = self.getGlobalVar(WarpWhistle.N106)
        shift = 0
        if N106_channels is not None and len(N106_channels):
            active_instruments = self.getDataForVoice(self.current_voices[0], WarpWhistle.INSTRUMENT)
            waveform = None
            for instrument in active_instruments:
                if hasattr(instrument, 'waveform'):
                    waveform = instrument.waveform
                    break

            shift = self.calculateN106OctaveShift(int(N106_channels), waveform)
        
        match = re.match(r'^(\[+)?([a-g](\+|\-)?)(.*)$', start_data['note'])
        start_data['note'] = match.group(2)
        start_data['append'] = match.group(4)

        # total amount we need to move
        slide_amount = self.getFrequency(start_data['note'], start_data['octave'] + shift) - self.getFrequency(end_data['note'], end_data['octave'] + shift)
                
        # song tempo
        tempo = self.getDataForVoice(self.current_voices[0], WarpWhistle.TEMPO)

        # @todo this is NTSC, need to add support for PAL - 50 FPS
        frames_per_second = 60

        # default to 16th note unless speed is specified
        slide_note_duration = 16 if start_data['speed'] is None else start_data['speed']

        # figure out the slide duration in seconds
        beats_per_second = float(tempo) / float(60)
        slides_per_second = (float(slide_note_duration) / float(4)) * beats_per_second
        slide_duration = float(1) / float(slides_per_second)

        frames_for_slide = int(math.floor(frames_per_second * slide_duration))

        steps = frames_for_slide
        distance_per_step = int(math.floor(slide_amount / frames_for_slide))
        remainder = slide_amount - steps * distance_per_step

        # print 'STEPS',steps
        # print 'DISTANCE_PER_STEP',distance_per_step
        # print 'REMAINDER',remainder

        increase_at = steps - remainder
        pitch_macro = []
        for x in range(0, steps):
            if x < increase_at:
                pitch_macro.append(str(distance_per_step))
                continue

            pitch_macro.append(str(distance_per_step + 1))

        pitch_macro = ' '.join(pitch_macro) + ' 0'
        instrument = Instrument({
            'pitch': pitch_macro
        })

        macro = instrument.getPitchMacro()

        # no longer need to slide
        self.setDataForVoices(self.current_voices, WarpWhistle.SLIDE, None)

        append_before = end_data['append']
        append_after =''
        match = re.match(r'(.*)(\](.*))', end_data['append'])
        if match:

            if match.group(1):
                append_before = match.group(1)

            if match.group(2):
                append_after = match.group(2)

        # print 'START',start_data
        # print 'END',end_data
        # print ""

        octave_diff = start_data['octave'] - end_data['octave']

        diff = Util.arrayDiff(self.current_voices, ['A', 'B', 'C'])

        smooth_start = ''
        smooth_end = ''
        if len(diff) == 0:
            smooth = self.getGlobalVar(WarpWhistle.SMOOTH)
        
            if smooth:
                smooth_start = ' SM '
                smooth_end = ' SMOF'

        return self.getOctaveShift(octave_diff) + smooth_start + macro + ' ' + start_data['note'] + append_before + ' EPOF' + smooth_end + append_after + ' ' + self.getOctaveShift(-octave_diff)

    def getOctaveShift(self, ticks):
        char = '<' if ticks < 0 else '>'
        return abs(ticks) * char

    def transposeNote(self, note, octave, amount, append):
        if append is None:
            append = ''

        start_data = self.getDataForVoice(self.current_voices[0], WarpWhistle.SLIDE)

        new_note = ''
        if amount == 0 or self.isNoiseChannel():
            if start_data:
                return self.slide(start_data, {'note': note, 'append': append, 'octave': octave})

            return note + append

        new_note_number = self.getNumberForNote()[note] + amount

        ticks = 0
        while new_note_number < 0:
            new_note += '< '
            ticks += 1
            new_note_number = new_note_number + 12

        while new_note_number > 11:
            ticks  -= 1
            new_note += '> '
            new_note_number = new_note_number - 12

        note_name = self.getNoteForNumber()[new_note_number]
        new_note += note_name
        new_note += append + ' ' + self.getOctaveShift(ticks)

        return new_note

    def processWord(self, word, next_word, prev_word):
        if not word:
            return word

        valid_commands = ['EPOF', 'ENOF', 'MPOF', 'PS', 'SDQR', 'SDOF', 'MHOF', 'SM', 'SMOF', 'EHOF']
        if word in valid_commands:
            if self.ignore:
                return ""
            
            return word
        
        # matches a voice declaration
        if re.match(r'[A-Z]{1,}$', word):
            
            self.current_voices = list(word)

            # processing everything, keep going
            if self.process_voice is None:
                return word
            
            # if we are processing a specific voice
            # and we are on that voice
            if self.process_voice in self.current_voices:
                self.ignore = False
                return self.process_voice
            
            # if we are processing a specific voice and we are not on that voice
            self.ignore = True
            return ""

        if self.ignore:
            return ""

        # slides for portamento
        match = re.match(r'^\/([0-9]+)?$', word)
        if match:

            # calculate the previous note
            prev_note = self.processWord(prev_word, None, None)

            # figure out what octave we are at now
            start_octave = self.getDataForVoice(self.current_voices[0], WarpWhistle.OCTAVE)

            self.setDataForVoices(self.current_voices, WarpWhistle.SLIDE, {'note': prev_note, 'octave': start_octave, 'speed': match.group(1)})

            return ''

        # matches a tempo declaration
        if re.match(r't\d+$', word):
            self.setDataForVoices(self.current_voices, WarpWhistle.TEMPO, int(word[1:]))
            return word

        # volume change
        if re.match(r'@v\d+$', word) and next != '=':
            self.setDataForVoices(self.current_voices, WarpWhistle.VOLUME, int(word[2:]))
            return word

        # timbre change
        if re.match(r'@@\d+$', word):
            self.setDataForVoices(self.current_voices, WarpWhistle.TIMBRE, int(word[2:]))
            return word

        # arpeggio change
        if re.match(r'EN\d+$', word):
            self.setDataForVoices(self.current_voices, WarpWhistle.ARPEGGIO, int(word[2:]))
            return word

        # pitch change
        if re.match(r'EP\d+$', word):
            self.setDataForVoices(self.current_voices, WarpWhistle.PITCH, int(word[2:]))
            return word

        # q change
        if re.match(r'q[0-8]$', word):
            self.setDataForVoices(self.current_voices, WarpWhistle.Q, int(word[1:]))
            return word

        # direct timbre
        if re.match(r'@\d+$', word):
            self.setDataForVoices(self.current_voices, WarpWhistle.TIMBRE, int(word[1:]))
            return word

        # explicit octave change (with o4 o3 etc)
        if re.match(r'o\d+$', word):
            self.setDataForVoices(self.current_voices, WarpWhistle.OCTAVE, int(word[1:]))
            return word

        # octave change with > or < or >>>
        if re.match(r'\>+|\<+$', word):
            direction = word[0]
            count = len(word)
            current_octave = self.getDataForVoice(self.current_voices[0], WarpWhistle.OCTAVE)

            if current_octave is None:
                current_octave = 0

            self.setDataForVoices(self.current_voices, WarpWhistle.OCTAVE, current_octave + (count if direction == '>' else -count))

            return word

        # dmc declaration
        match = re.match(r'(\{\s{0,})?(\'|\")(.*\.dmc)(\2)\s{0,},', word)
        if match:
            mmlx_dir = self.options['start'] if os.path.isdir(self.options['start']) else os.path.dirname(self.options['start'])
            new_path = os.path.join(mmlx_dir, match.group(3))
            new_word = ''

            if match.group(1):
                new_word += match.group(1)

            new_word += match.group(2) + new_path + match.group(2) + ','

            return new_word

        # rewrite special voices for mmlx such as c4 or G+,4^8
        # to use this put the line X-ABSOLUTE-NOTES at the top of your mmlx file
        match = re.match(r'(\[+)?([A-Ga-g]{1})(\+|\-)?(\d{1,2})?(,(\d+\.?)(\^[0-9\^]+)?)?([\]\d]+)?$', word)
        if match and self.getGlobalVar(WarpWhistle.ABSOLUTE_NOTES):
            is_noise_channel = self.current_voices[0] == 'D'

            if is_noise_channel and not "," in word:
                return word

            new_word = ""

            octave = match.group(4) if not is_noise_channel else 0

            current_octave = self.getDataForVoice(self.current_voices[0], WarpWhistle.OCTAVE)

            if current_octave is None and not is_noise_channel:
                new_word += 'o' + octave + ' '
            elif not is_noise_channel and octave and int(octave) != current_octave:
                new_word += self.moveToOctave(int(octave), current_octave) + ' '

            if octave:
                self.setDataForVoices(self.current_voices, WarpWhistle.OCTAVE, int(octave))
                current_octave = int(octave)

            # [[[
            if match.group(1):
                new_word += match.group(1)

            note = ""

            # note
            note += match.group(2).lower()

            # accidental
            if match.group(3):
                note += match.group(3)

            append = ""

            # tack on the note length
            if match.group(6):
                append += match.group(6)

            # tack on any ties (such as ^8^16)
            if match.group(7):
                append +=  match.group(7)

            # tack on the final repeat value if it is present (]4)
            if match.group(8):
                append += match.group(8)

            new_word += self.transposeNote(note, current_octave, self.getGlobalVar(WarpWhistle.TRANSPOSE), append)

            return new_word

        # regular note
        match = re.match(r'(\[+)?([a-g]{1}(\+|\-)?)([\.0-9\^]+)?([\]\d+]+)?$', word)
        if match:
            if "," in word and not self.getGlobalVar(WarpWhistle.ABSOLUTE_NOTES):
                raise Exception('In order to use absolute notes you have to specify X-ABSOLUTE-NOTES')

            current_octave = self.getDataForVoice(self.current_voices[0], WarpWhistle.OCTAVE)

            new_note = ""
            if match.group(1):
                new_note += match.group(1)

            append = ''
            if match.group(4):
                append = match.group(4)

            if match.group(5):
                append += match.group(5)

            new_note += self.transposeNote(match.group(2), current_octave, self.getGlobalVar(WarpWhistle.TRANSPOSE), append)

            return new_note

        # instrument
        match = re.match(r'^(\[+)?(\+)?@([a-zA-Z0-9-_]+)([\]\d+]+)?$', word)
        if match:

            new_word = ''

            # special case if you do @end you can end the currently active instruments
            if match.group(3) == 'end':
                active_instruments = self.getDataForVoice(self.current_voices[0], WarpWhistle.INSTRUMENT)

                for active_instrument in active_instruments:
                    new_word += active_instrument.end(self)

                self.setDataForVoices(self.current_voices, WarpWhistle.INSTRUMENT, [])
                return new_word

            # not a valid instrument
            if not match.group(3) in self.instruments:
                return word

            new_instrument = self.instruments[match.group(3)]
            
            if 'O' in self.current_voices and hasattr(new_instrument, 'timbre'):
                raise Exception('VRC6 sawtooth (voice O) does not support timbre attribute')

            chip = new_instrument.getChip()
            diff = []
            if chip is not None:
                diff = Util.arrayDiff(self.current_voices, self.getVoicesForChip(chip).values())
            
            if len(diff):
                diff.sort()
                words = ('voice', 'does') if len(diff) == 1 else ('voices', 'do')
                raise Exception(words[0] + ' ' + ', '.join(diff) + ' ' + words[1] + ' not support instruments using chip: ' + chip)

            active_instruments = self.getDataForVoice(self.current_voices[0], WarpWhistle.INSTRUMENT)

            if active_instruments is None:
                active_instruments = []

            if match.group(1):
                new_word += match.group(1)

            if len(active_instruments) and not match.group(2):
                for active_instrument in active_instruments:
                    new_word += active_instrument.end(self)

                active_instruments = []

            active_instruments.append(new_instrument)
            self.setDataForVoices(self.current_voices, WarpWhistle.INSTRUMENT, active_instruments)

            new_word += new_instrument.start(self)

            if match.group(4):
                new_word += match.group(4)

            return new_word

        if self.isUndefinedVariable(word):
            raise Exception('variable ' + word + ' is undefined')

        # print "PROCESS:",word
        # print "PREV:",prev_word
        # print "NEXT:",next_word
        # print ""
        return word

    def processLine(self, line):
        self.ignore = False

        words = line.split(' ')
        new_words = []

        for key, word in enumerate(words):
            next_word = None
            prev_word = None

            if len(words) > key + 1:
                next_word = words[key + 1]

            if len(words) > key - 1:
                prev_word = words[key - 1]

            new_words.append(self.processWord(word, next_word, prev_word))

        return ' '.join(new_words)
    
    def findVoices(self, content):
        matches = re.findall(r'^([A-Z]{1,}) ', content, re.MULTILINE)

        voices = []
        for match in matches:
            new_voices = list(match)
            for voice in new_voices:
                if not voice in voices:
                    voices.append(voice)
        
        return voices

    def process(self, content):
        self.logger.log('- stripping comments', True)
        content = self.stripComments(content)

        self.logger.log('- proccessing imports', True)
        content = self.processImports(content)

        self.logger.log('- stripping comments again', True)
        content = self.stripComments(content)

        self.logger.log('- parsing variables', True)
        content = self.processVariables(content)

        self.logger.log('- applying variables', True)
        content = self.replaceVariables(content)

        self.logger.log('- parsing instruments', True)
        content = self.processInstruments(content)

        self.logger.log('- collapsing spaces', True)
        content = self.collapseSpaces(content)
        
        self.logger.log('- processing expansion voices', True)
        content = self.processExpansionVoices(content)

        self.voices = self.findVoices(content)
        content = self.renderTempo(content)

        content = self.renderExpansionChips(content)

        if not self.first_run:
            if self.voices_to_process is None:
                self.voices_to_process = self.voices
            
            if len(self.voices_to_process):
                self.process_voice = self.voices_to_process.pop(0)
            
        if self.process_voice:
            self.logger.log('processing voice: ' + self.process_voice, True)

        lines = content.split('\n')
        new_lines = []
        for line in lines:
            new_lines.append(self.processLine(line))

        content = '\n'.join(new_lines)

        content = self.renderInstruments(content)

        self.logger.log('- replacing unneccessary octave shifts', True)
        levels = math.ceil(abs(float(self.getGlobalVar(WarpWhistle.TRANSPOSE))) / float(12))
        for x in range(0, int(levels) + 1):
            content = self.collapseSpaces(content)
            content = content.replace('> <', '').replace('< >', '')

        content = self.collapseSpaces(content)

        self.logger.log('- removing blank lines', True)
        content = self.removeBlankLines(content)

        self.first_run = False

        return content

    def isPlaying(self):
        if self.first_run:
            return True

        if not self.options['separate_voices']:
            return False
        
        return self.voices_to_process is None or len(self.voices_to_process) != 0

    def play(self):
        counter = self.getGlobalVar(WarpWhistle.COUNTER) or 0
        Instrument.reset(counter)
        self.reset()
        return (self.process(self.content), self.process_voice)



#!/usr/bin/env python

from distutils.core import setup

setup(
    name = 'mmlx',
    version = '1.0.2',
    description = 'Nintendo NES chiptune programming language',
    long_description = 'MMLX stands for MML eXtended.  It is an extended version of Music Macro Language that generates MML and NSF files for Nintendo NES music composing',
    author = 'Craig Campbell',
    author_email = 'iamcraigcampbell@gmail.com',
    url = 'https://github.com/ccampbell/mmlx',
    download_url = 'https://github.com/ccampbell/mmlx/zipball/1.0.2',
    license = 'Apache Software License',
    packages = ['mmlxlib'],
    package_data = {'mmlxlib': ['nes_include/ppmck.asm', 'nes_include/ppmck/*']},
    scripts = ['bin/mmlx', 'bin/ppmckc', 'bin/nesasm']
)
